{-# LANGUAGE GADTs                 #-}
{-# LANGUAGE DataKinds             #-}
{-# LANGUAGE TypeFamilies          #-}
{-# LANGUAGE TypeOperators         #-}
{-# LANGUAGE StandaloneDeriving    #-}
{-# LANGUAGE ScopedTypeVariables   #-}
{-# LANGUAGE UndecidableInstances  #-}
{-# LANGUAGE TemplateHaskell       #-}
{-# LANGUAGE QuasiQuotes           #-}
{-# LANGUAGE PolyKinds             #-}
{-# LANGUAGE UnicodeSyntax         #-}
{-# LANGUAGE RebindableSyntax      #-}

-- | Defines the boolean type and logical operators
module Data.Bool where

import Data.Kind      (Type)
import Data.TypeClass (Eq(..), Enum, Show)
import GHC.Types      (Bool(..))
import Data.Singletons.TH


class IfThenElse b where
  ifThenElse ∷ b -> a -> a -> a


singletons [d|

  -- | Boolean Type
  data 𝔹 ∷ Type where
    T ∷ 𝔹 -- ^ True
    F ∷ 𝔹 -- ^ False
    deriving (Eq, Show)


  -- * Logical operators

  -- | Conjuction
  (∧) ∷ 𝔹 → 𝔹 → 𝔹
  T ∧ T = T
  _ ∧ F = F
  F ∧ _ = F
  
  -- | Disjuction
  (∨) ∷ 𝔹 → 𝔹 → 𝔹
  T ∨ _ = T
  _ ∨ T = T
  F ∨ F = F

  -- | Negation
  (¬) ∷ 𝔹 → 𝔹
  (¬) T = F
  (¬) F = T

  |]

-- | Values equality
(≡) ∷ Eq a ⇒ a → a → 𝔹
x ≡ y = case x == y of
  True  → T
  False → F

instance IfThenElse 𝔹 where
  ifThenElse T x y = x
  ifThenElse F x y = y

