{-# LANGUAGE GADTs                 #-}
{-# LANGUAGE DataKinds             #-}
{-# LANGUAGE TypeFamilies          #-}
{-# LANGUAGE TypeOperators         #-}
{-# LANGUAGE StandaloneDeriving    #-}
{-# LANGUAGE ScopedTypeVariables   #-}
{-# LANGUAGE NoImplicitPrelude     #-}
{-# LANGUAGE UndecidableInstances  #-}
{-# LANGUAGE TemplateHaskell       #-}
{-# LANGUAGE QuasiQuotes           #-}
{-# LANGUAGE PolyKinds             #-}
{-# LANGUAGE UnicodeSyntax         #-}

-- | Defines commonly used functions and combinators
module Data.Function where

import Data.Singletons.TH

singletons [d|

  -- | Identity function
  id ∷ ∀ a. a → a
  id x = x

  -- | Constant function
  const ∷ ∀ a b. a → b → a
  const k _ = k

  -- | Flip the arguments of a function
  flip ∷ ∀ a b c. (a → b → c) → (b → a → c)
  flip f x y = f y x

  -- | Undefined
  (⊥) ∷ ∀ a. a
  (⊥) = (⊥)

  infixr 9 ∘

  -- | Function composition operator
  (∘) :: ∀ a b c. (b → c) → (a → b) → a → c
  f ∘ g = \x → f (g x)

  |]

infixr 0 $

-- | Function application operator
($) ∷ ∀ a b. (a → b) → a → b
f $ x = f x
